#!/usr/bin/env python


''' RDI-specific terminal (BB, WH) for LADCP or moored
ADCP deployment and recovery. '''

from tk_terminal import Tk_terminal, Timeout

import time, termios
import sys, os, select, signal
import shutil, re
from threading import Thread
import Queue
import commands

from Tkinter import *
import tkFileDialog, tkMessageBox
from tkSimpleDialog import askinteger, askstring
import Pmw


## adapted from sonterm for WH, starting 2003/06/08
## adapted at sea (JH) for 150KHz BB 2003/06/19.  needs integration
## more changes for shipboard OS use. EF 2003/12/13
## more changes for LADCP use. EF 2004/05/20
## changes for python2.5 and keyspan2.5 by AJH & TKC 8/15/2008 for macbook aoraki2

## For Mac OSX (darwin): http://fink.sourceforge.net/pdb/package.php/lrzsz

rdi_baud_codes = {300:0, 1200:1, 2400:2,
                  4800:3, 9600:4, 19200:5,
                  38400:6, 57600:7, 115200:8}

fmt = '%Y/%m/%d  %H:%M:%S'
YM_retry = re.compile("Retry (\d+): TIMEOUT")
dev_name = '/dev/tty.KeySerial1'

def time_stamp():
   return time.strftime(fmt, time.gmtime())

class terminal(Tk_terminal):
   def __init__(self, master = None, device = dev_name,
                      baud = 9600, Loggers = None,
                      cmd_filename = 'ladcp.cmd',
                      prefix = 'rdi',
                      backup = '',
                      stacast = '000_01',
                      datafile_ext = '.dat'):
      Tk_terminal.__init__(self, master = master,
                     device = device, baud = baud)
      self.default_baud = baud
      self.cmd_filename = cmd_filename
      self.prefix = prefix
      self.backup = backup
      self.stacastSV = StringVar()
      self.stacastSV.set(stacast)
      if not datafile_ext.startswith('.'):
         datafile_ext = '.%s' % datafile_ext
      self.datafile_ext = datafile_ext
      self.Loggers = Loggers
      self.type = ''         # After wakeup: BB or WH
      self.prefixline = Pmw.LabeledWidget(self.statusframe,
                                          labelpos = 'w',
                                          label_text = 'Prefix:')
      self.prefixline.pack(side = LEFT, anchor = W, padx = 10)
      Label(self.prefixline.interior(),
            relief = SUNKEN,
            text = self.prefix).pack()
      self.stacastline = Pmw.LabeledWidget(self.statusframe,
                                          labelpos = 'w',
                                          label_text = 'Station_Cast:')
      self.stacastline.pack(side = LEFT, anchor = W, padx = 10)
      self.stacastentry = Entry(self.stacastline.interior(),
            #relief = SUNKEN,
            width = 10,
            bg = 'white',
            textvariable = self.stacastSV)
      self.stacastentry.pack()

   def add_Loggers(self, Loggers):
      self.Loggers = Loggers

   def start_logging(self):
      if not self.Loggers:
         return
      self.Loggers.start_logging()

   def stop_logging(self):
      if not self.Loggers:
         return
      self.Loggers.check_stop_logging()

   def make_filename(self, ext):
      return "%s%s%s" % (self.prefix, self.stacastSV.get(), ext)

   def insert(self, msg):
      ''' Insert a comment as if it came from the instrument.
          Add a leading "#" to each line.
      '''
      linelist = msg.splitlines()
      msg1 = "\n#%s\n" % "\n#".join(linelist)
      self.new_input.put(msg1)

   def start_listening(self):
      self.stop_logging()
      Tk_terminal.start_listening(self)

   def set_clock(self):
      self.wakeup()
      DateTime = time.strftime('%y%m%d%H%M%S', time.gmtime())
      self.send_commands(('TS%s' % DateTime, ))

   def show_time_ok(self):
      try:
         t_pc = time.gmtime()
         self.send_commands(('TS?',))
      except:
         self.wakeup()
         t_pc = time.gmtime()
         self.send_commands(('TS?',)) #tkc added back
      TSline = self.display.get_line_with('TS').strip()
      TSstring = TSline.split()[2]
      t_adcp = time.strptime(TSstring, '%y/%m/%d,%H:%M:%S')
      # force the DST field to be zero
      t_adcp = t_adcp[0:8] + (0,)
      dt = time.mktime(t_pc) - time.mktime(t_adcp)
      msg = ('  PC time:   %s\n'
             'ADCP time:   %s\n'
             'PC-ADCP:     %.0f' % (time.strftime(fmt, t_pc),
                                      time.strftime(fmt, t_adcp),
                                      dt))
      self.insert(msg)
      #tkMessageBox.showinfo(message = msg)  # Ugly!
      Pmw.MessageDialog(title = 'Time', message_text = msg,
                        buttons = ('OK',)).activate()

   def showconfig(self):
      commands = ('RA','RS','RB0','PS3','B?','C?','E?','P?','T?', 'W?')
      try:
         self.send_commands(commands)
      except:
         self.wakeup()
         self.send_commands(commands)

   def wakeup(self):
      self.change_baud(self.default_baud)
      self.start_listening()
      self.display.mark()
      self.display.mark(name = 'Banner')
      self.send_break()
      self.waitfor('>', 1.5)
      banner = self.display.get_after_mark(name = 'Banner')
      if banner.find('WorkHorse') > -1:
         self.type = 'WH'
      elif banner.find('Broadband') > -1:
         self.type = 'BB'
      else:
         self.type = 'Unrecognized'

   def wakeup_showtime(self):
      self.wakeup()
      self.show_time_ok()

   def sleep(self):
      self.change_baud(self.default_baud)
      self.start_listening()
      self.display.mark()
      self.send_break()
      self.waitfor('>', 3)
      self.stream.write('CZ' + '\r')
      self.stream.flush()
      if self.type == 'BB':
         self.waitfor('[POWERING DOWN .....]', 2)
      else:
         self.waitfor('Powering Down', 2)

   def eraserecorder(self):
      self.change_baud(default_baud)
      self.start_listening()
      self.display.mark()
      self.send_break()
      self.waitfor('>', 3)
      self.stream.write('RE ErAsE' + '\r')
      self.stream.flush()
      self.waitfor('>', 2)

   def send_commands(self, commands):
      self.start_listening()
      self.display.mark()
      for cmd in commands:
         self.stream.write(cmd.rstrip() + '\r')
         self.stream.flush()
         self.waitfor('>')

   def send_setup(self):
      fmt = '%Y/%m/%d  %H:%M:%S'
      self.wakeup() # Make sure it is listening.
      fn = self.cmd_filename
      self.insert("Sending command file: %s" % fn)
      try:
         lines = open(fn, 'r').readlines()
         cmds = [line.strip() for line in lines if len(line) > 1]
         self.send_commands(cmds)
         self.stream.write('CK' + '\r')
         self.stream.flush()
         self.waitfor('[Parameters saved as USER defaults]', 2)
         self.stream.write('CS' + '\r')
         self.stream.flush()

      except:
         tkMessageBox.showerror(message = "Can't send file %s" % fn)
         self.ask_send_setup()
      time.sleep(1)
      self.insert("Data collection started, %s\n" % time_stamp())
      logfilename = self.make_filename(".log")
      self.append_to_file(logfilename)
      self.insert("Deployment logfile written to %s" % logfilename)



   def ask_send_setup(self):
      fn = tkFileDialog.askopenfilename(initialfile = self.cmd_filename,
                     initialdir = os.getcwd(),
                     filetypes = (('Command', '*.cmd'), ('All', '*')),
                     parent = self.Frame,
                     title = 'Command file')
      if fn == '': return
      self.cmd_filename = fn
      self.send_setup()



   # This may not be useful; it is supposed to
   # get a single ensemble in hex-ascii mode.
   def start_ascii(self):
      self.wakeup()
      #self.set_clock()
      self.stream.write('CF01010\r')
      self.stream.flush()
      termios.tcdrain(self.fd)
      self.stream.write('CS\r')
      self.stream.flush()
      termios.tcdrain(self.fd)
      self.waitfor('>')


   # ping at will
   def start_binary(self, cmdlist = None):
      self.wakeup()
      self.set_clock()
      if cmdlist:
         self.send_commands(cmdlist)
      self.stream.write('CF11110\r') # serial out, no recorder
      self.stream.flush()
      termios.tcdrain(self.fd)
      self.change_all_baud()  ### might not want this here in general
      self.mark()
      self.stream.write('CS\r')
      self.stream.flush()
      termios.tcdrain(self.fd)
      self.waitfor('CS\r\n')
      self.stop_listening()
      self.start_logging()

   def change_all_baud(self, baud = 115200):
      self.send_commands([''])
      self.stream.write('CB%d11\r' % rdi_baud_codes[baud])
      self.stream.flush()
      termios.tcdrain(self.fd)
      self.waitfor('>', 2)
      self.change_baud(baud)

   def list_recorder(self):
      if self.type == 'BB':
         cmds = ('RA', 'RS')
      else:
         cmds = ('RA', 'RS', 'RF', 'RR')
      try:
         self.send_commands(cmds)
      except Timeout:
         self.wakeup()
         self.send_commands(cmds)

   def download_selected(self):
      try:
         ind1 = self.display.search('Bytes used on device',
                                    END, backwards = 1)
         ind0 = self.display.search('Volume serial number',
                                    END, backwards = 1)
         base_line = int(split(ind0, '.')[0])
         ind = self.display.index('sel.first')
         sel_line = int(split(ind, '.')[0])
         n = sel_line - base_line
         self.ymodem_download(n)

      except:
         pass

   def find_filename(self):
      '''For BB instrument, get name of last file downloaded.'''
      line = self.display.get_line_with('Receiving:').strip()
      filename = line.split()[1]
      print filename
      return filename

   def find_number_recorded(self):
      try:
         self.send_commands(('RA',))
      except Timeout:
         self.wakeup()
         self.send_commands(('RA',))
      if self.type == 'BB':
         ii = 2
      else:
         ii = 1

      line = self.display.get_lines_from('RA')
      print line.split()
      return int(line.split()[ii])


   def download(self):
      if not self.type:
         self.wakeup()
      nrec = self.find_number_recorded()
      ndown = askinteger('Download',
              'File to download, %d to %d' % (1, nrec),
              initialvalue = nrec,
              minvalue = 1,
              maxvalue = nrec)
      if ndown is None:
         return
      self.ymodem_download(ndown)
      if self.canceled:
         return
      fn0 = self.find_filename()
      os.system('touch %s' % fn0)
      fn = self.make_filename(self.datafile_ext)
      fn1 = None
      while fn1 is None:
         fn1 = askstring('Rename', 'Rename %s to:' % fn0,
                 initialvalue = fn)
         if fn1 is None:
            break
         if os.path.exists(fn1):
            tkMessageBox.showerror(message = "File %s already exists" % fn1)
            fn1 = None
      if fn1:
         os.rename(fn0, fn1)
         self.insert("File written as %s" % fn1)
      else:
         fn1 = fn0
      os.chmod(fn1, 0444) # Read-only.
      if self.backup:
         try:
            shutil.copy2(fn1, os.path.join(self.backup, fn1))
            self.insert("File %s backed up to %s" % (fn1, self.backup))
         except:
            tkMessageBox.showerror(message =
                              "Backup to %s failed." % self.backup)
      logfilename = self.make_filename(".log")
      self.append_to_file(logfilename)
      self.insert("Recovery logfile appended to %s" % logfilename)
      self.sleep()

   ## Attempts to use update and listen from tk_terminal for
   ## pipe reading here failed; although it seems like it must
   ## be possible to make a single update/listen pair that would
   ## work with pipes and serial ports, I have not managed to do
   ## it successfully yet.
   def update2(self):    # Used by read_pipe.
      try:
         while 1:
            cc = self.new_input.get_nowait()
            self.display.append(cc)
      except Queue.Empty:
         pass
      if self.listening2:
         self.Frame.after(50, self.update2)


   def listen2(self, pipename):   # Run in a separate thread by read_pipe.
      fd = os.open(pipename, os.O_CREAT | os.O_RDONLY | os.O_NONBLOCK)
      while self.listening2:
         # Select provides a timeout so self.listening2 is checked.
         rd, wr, er = select.select([fd], [], [], 0.1)
         if fd in rd:
            cc = os.read(fd, 500)
            cc = cc.replace('\r', '\n')
            if cc:
               self.new_input.put(cc)  # via queue to main thread
               if self.save:
                  self.outfile.write(cc)
         time.sleep(0.001) # Seems to be necessary, contrary to Grayson.
      os.close(fd)

   def read_pipe(self, pipename):
      '''Read the output from the rb command (ymodem download)
         until the word 'complete' signals the end of the
         transfer.
      '''
      self.listening2 = 1
      self.thread2 = Thread(target = self.listen2, args = (pipename,))
      self.thread2.setDaemon(0)
      self.display.mark()
      self.thread2.start()
      self.update2()
      while self.listening2:
         self.master.update()
         try:
            self.waitfor('complete', 0.05)
            self.listening2 = 0
            while self.thread2.isAlive():
               time.sleep(0.05)
         except Timeout:
            pass

   def abort_download(self):
      self.listening2 = 0
      try:
         rb_pid = commands.getoutput('ps -C rb -o pid --noheader').split()[0]
         os.kill(int(rb_pid), signal.SIGTERM)
      except OSError, msg:
         print msg
      except:
         pass
      time.sleep(0.5)
      self.Frame.update()
      self.wakeup()
      self.insert("DOWNLOAD CANCELED")
      self.canceled = 1


   def ymodem_download(self, filenum = None):
      # filenum is None for all files (WH only); otherwise file number
      #            directory selection; for now it is in current directory
      self.wakeup()
      pipename = os.tempnam()
      os.mkfifo(pipename)
      print "A"
      mybaud_rate = 115200
      self.change_all_baud(mybaud_rate)
      print "B"
      if filenum:
         self.stream.write('RY%d\r' % filenum)
      else:
         self.stream.write('RY\r' % filenum)
      self.stream.flush()
      print "C type = %s" % self.type

      if self.type == 'WH':
         self.waitfor('now')
      self.stop_listening()
      d = self.get_device()
      self.canceled = 0
      self.mark(name = "YmodemDownload")

      mycommand = "(/sw/bin/lrb  --rename < %s >> %s 2>%s &);sleep 2;stty -f %s speed %d" %  (
                         d, d, pipename,d,mybaud_rate)
      print mycommand
      os.system(mycommand)

#       if sys.platform == 'darwin':
#         os.system("stty -f %s speed 115200 ; (/sw/bin/lrb  --rename < %s >> %s 2>%s &)" %
#                   (d, d, d, pipename))
#       else:
#          os.system("stty -F %s 115200; (rb  --rename < %s >> %s 2>%s &)" %
#                      (d, d, d, pipename))

      self.set_status("Ymodem download on %s at %d baud."
                              % (self.get_device(),mybaud_rate))
      b_cancel = Button(self.toprow, text = 'Cancel Download',
                        command = self.abort_download)
      b_cancel.pack(side = RIGHT, expand = NO, fill = NONE)
      self.menubar.disableall()
      self.Frame.update()
      self.read_pipe(pipename)
      self.menubar.enableall()
      b_cancel.destroy()
      self.Frame.update()
      os.unlink(pipename)
      YMlog = self.display.get_after_mark("YmodemDownload")
      nretry = 0
      for m in YM_retry.finditer(YMlog):
         nretry = nretry + 1 + int(m.group(1))
      self.insert("ymodem download number of retries: %d" % nretry)
      self.start_listening()
      self.change_all_baud(self.default_baud)
      self.stop_listening()
      os.system("stty -f %s speed 9600" % d)   # NEW AJH
      self.start_listening()
      
   def append_to_file(self, fn):
      try:
         f = open(fn, 'a')
         f.write(self.display.get())
         f.close()
      except:
         tkMessageBox.showerror(message = "Can't write to file %s" % fn)

   def start_deploy_recover(self):
      self.clear()
      self.insert("***************************** %s\n" % time_stamp())
      self.wakeup_showtime()
      self.list_recorder()

   def make_menu(self, master):
      Tk_terminal.make_menu(self, master)
      mb = self.menubar
      mb.deletemenuitems('Command', 0)

      mb.addmenuitem('Command', 'command', '',
                     label = 'Wakeup',
                     command = self.wakeup)
      mb.addmenuitem('Command', 'command', '',
                     label = 'ZZZZ (go to sleep)',
                     command = self.sleep)
      mb.addmenuitem('Command', 'command', '',
                     label = 'Set Clock',
                     command = self.set_clock)
      mb.addmenuitem('Command', 'command', '',
                     label = 'Send Setup',
                     command = self.ask_send_setup)
      mb.addmenuitem('Command', 'command', '',
                     label = 'Show Config',
                     command = self.showconfig)

      mb.addmenuitem('Command', 'command', '',
                     label = '115200 Baud',
                     command = self.change_all_baud)
      mb.addmenuitem('Command', 'command', '',
                     label = 'Start Binary',
                     command = self.start_binary)
      #mb.addmenuitem('Command', 'command', '',
      #               label = 'Start ASCII (1 ens)',
      #               command = self.start_ascii)
      mb.addmenuitem('Command', 'command', '',
                     label = 'List Recorder Directory',
                     command = self.list_recorder)
      #mb.addmenuitem('Command', 'command', '',
      #               label = 'Download all files',
      #               command = self.ymodem_download)

      ## disabled for now 2003/06/18
      #mb.addmenuitem('Command', 'command', '',
      #               label = 'Download marked file',
      #               #command = self.download_marked)
      #               command = self.download_selected)
# disabled for now 2005/06/04
# until I can add a gui confirm box
#       mb.addmenuitem('Command', 'command', '',
#                      label = 'Erase Recorder NOW',
#                      command = self.eraserecorder)


      mb.addmenu('Deploy', '')
      mb.addmenuitem('Deploy', 'command', '',
                     label = 'Deployment Initialization',
                     command = self.start_deploy_recover)
      mb.addmenuitem('Deploy', 'command', '',
                     label = 'Set Clock',
                     command = self.set_clock)
      #mb.addmenuitem('Deploy', 'command', '',
      #               label = 'List Recorder Directory',
      #               command = self.list_recorder)
      mb.addmenuitem('Deploy', 'command', '',
                     label = 'Send Setup',
                     command = self.ask_send_setup)
      mb.addmenuitem('Deploy', 'command', '',
                     label = 'Disconnect',
                     command = self.stop_listening)


      mb.addmenu('Recover', '')
      mb.addmenuitem('Recover', 'command', '',
                     label = 'Recovery Initialization',
                     command = self.start_deploy_recover)
      #mb.addmenuitem('Recover', 'command', '',
      #               label = 'Wakeup_showtime',
      #               command = self.wakeup_showtime)
      #mb.addmenuitem('Recover', 'command', '',
      #               label = 'List Recorder Directory',
      #               command = self.list_recorder)
      #mb.addmenuitem('Recover', 'command', '',
      #               label = 'Clear Screen',
      #               command = self.clear)
      #mb.addmenuitem('Recover', 'command', '',
      #               label = 'Show Config',
      #               command = self.showconfig)
      #mb.addmenuitem('Recover', 'command', '',
      #               label = 'Save Previous',
      #               command = self.ask_write_file)
      mb.addmenuitem('Recover', 'command', '',
                     label = 'Download',
                     command = self.download)
      #mb.addmenuitem('Recover', 'command', '',
      #               label = 'ZZZZ (go to sleep)',
      #               command = self.sleep)

if __name__ == '__main__':
   import sys
   from optparse import OptionParser
   op = OptionParser()
   op.add_option('-d', '--device', dest = "device",
                  help = ('DEVICE is the full path of serial device;' +
                            ' default is %s' % dev_name),
                  default = dev_name)
   op.add_option('-p', '--prefix', dest = "prefix",
                  help = 'prefix of default name for downloaded file',
                  default = 'N')
   op.add_option('-b', '--backup', dest = "backup",
                  help = 'directory for automatic backup',
                  default = '')
   op.add_option('', '--baud', dest = "baud", type = 'int',
                  help = 'default baud rate',
                  default = 9600)
   op.add_option('-s', '--stacast', dest = "stacast",
                  help = 'Station_Cast, e.g., 057_01, for file naming',
                  default = '000_01')
   op.add_option('-e', '--datafile_ext', dest = "datafile_ext",
                  help = 'extension for downloaded data files',
                  default = '000')
   op.add_option('-c', '--command', dest = "cmd_filename",
                  help = "Setup command filename",
                  default = 'ladcp.cmd')

   o, a = op.parse_args()

   R = terminal(device = o.device, master = None,
                  baud = o.baud,
                  prefix = o.prefix,
                  backup = o.backup,
                  stacast = o.stacast,
                  datafile_ext = o.datafile_ext,
                  cmd_filename = o.cmd_filename)
   logfilename = 'rditerm_%s.log' % os.path.split(o.device)[-1]
   print "Saving terminal IO to %s." % logfilename
   R.begin_save(logfilename)
   R.master.mainloop()



